#include "vector_tile_processor.hpp"
#include "vector_tile_processor.ipp"
#include "vector_tile_backend_pbf.hpp"
template class mapnik::vector_tile_impl::processor<mapnik::vector_tile_impl::backend_pbf>;
