#ifndef __MAPNIK_VECTOR_TILE_CONFIG_H__
#define __MAPNIK_VECTOR_TILE_CONFIG_H__

#if !defined(MAPNIK_VECTOR_TILE_LIBRARY)
#define MAPNIK_VECTOR_INLINE inline
#else
#define MAPNIK_VECTOR_INLINE
#endif


#endif // __MAPNIK_VECTOR_TILE_CONFIG_H__